@@ -16,21 +16,25 @@ as well as some convenience functions.
 :authors: M. Marin (KTH)
 :date:   12.10.2020
 """
-
+import contextlib
 import socket
 import socketserver
 import sys
 import os
 import signal
+import threading
 import time
+from dataclasses import dataclass
+from itertools import chain
 from subprocess import Popen, PIPE
-from threading import Thread
-from multiprocessing import Lock
-from queue import Queue
+from multiprocessing import Process, Lock
+from queue import Queue, Empty
 from ..utils.helpers import Singleton
+from ..utils.execution_resources import persistent_process_semaphore
 from .execution_manager_base import ExecutionManagerBase
 
 if sys.platform == "win32":
+    import ctypes
     from subprocess import CREATE_NEW_PROCESS_GROUP, CREATE_NO_WINDOW
 
 
@@ -41,10 +45,11 @@ class PersistentManagerBase:
             args (list): the arguments to launch the persistent process
         """
         self._args = args
-        self._command_buffer = []
         self._server_address = None
         self._msg_queue = Queue()
         self.command_successful = False
+        self._is_running_lock = Lock()
+        self._is_running = True
         self._kwargs = dict(stdin=PIPE, stdout=PIPE, stderr=PIPE)
         if sys.platform == "win32":
             self._kwargs["creationflags"] = CREATE_NEW_PROCESS_GROUP | CREATE_NO_WINDOW
@@ -53,6 +58,7 @@ class PersistentManagerBase:
         self._lock = Lock()
         self._persistent = None
         self._start_persistent()
+        self._wait()
 
     @property
     def language(self):
@@ -63,6 +69,19 @@ class PersistentManagerBase:
         """
         raise NotImplementedError()
 
+    def set_running_until_completion(self, running):
+        """Sets the 'running until completion' state.
+
+        Args:
+            running (bool): True if the manager is running until completion, False otherwise
+        """
+        with self._is_running_lock:
+            self._is_running = running
+
+    def is_running_until_completion(self):
+        with self._is_running_lock:
+            return self._is_running
+
     def _init_args(self):
         """Returns init args for Popen.
 
@@ -75,9 +94,21 @@ class PersistentManagerBase:
         raise NotImplementedError()
 
     @staticmethod
-    def _sentinel_command(host, port):
-        """Returns a command in the underlying language, that sends a sentinel to a socket listening on given
-        host/port.
+    def _catch_exception_command(cmd):
+        """Returns an equivalent command that also recorders if an exception is raised.
+
+        Args:
+            cmd (str)
+
+        Returns:
+            str
+        """
+        raise NotImplementedError()
+
+    @staticmethod
+    def _ping_command(host, port):
+        """Returns a command in the underlying language that connects to a socket listening on given host/port,
+        and sends "ok" if no exception has been recorded.
 
         Used to synchronize with the persistent process.
 
@@ -97,55 +128,66 @@ class PersistentManagerBase:
             self._server_address = s.server_address
         self.command_successful = False
         self._persistent = Popen(self._args + self._init_args(), **self._kwargs)
-        Thread(target=self._log_stdout, daemon=True).start()
-        Thread(target=self._log_stderr, daemon=True).start()
+        threading.Thread(target=self._log_stdout, daemon=True).start()
+        threading.Thread(target=self._log_stderr, daemon=True).start()
 
     def _log_stdout(self):
         """Puts stdout from the process into the queue (it will be consumed by issue_command())."""
         for line in iter(self._persistent.stdout.readline, b''):
-            data = line.decode("UTF8", "replace").strip()
+            data = line.decode("UTF8", "replace").rstrip()
             self._msg_queue.put(dict(type="stdout", data=data))
 
     def _log_stderr(self):
         """Puts stderr from the process into the queue (it will be consumed by issue_command())."""
         for line in iter(self._persistent.stderr.readline, b''):
-            data = line.decode("UTF8", "replace").strip()
+            data = line.decode("UTF8", "replace").rstrip()
             self._msg_queue.put(dict(type="stderr", data=data))
 
-    def _is_complete(self, cmd):
+    def make_complete_command(self, cmd):
+        lines = cmd.splitlines()
+        cmd = os.linesep.join(lines)
+        if len(lines) == 1:
+            cmd += os.linesep
         result = self._communicate("is_complete", cmd)
-        return result.strip() == "true"
+        if result.strip() != "true":
+            return None
+        return cmd
 
-    def issue_command(self, cmd, add_history=False):
-        """Issues cmd to the persistent process and returns an iterator of stdout and stderr messages.
+    def drain_queue(self):
+        while True:
+            try:
+                yield self._msg_queue.get(timeout=0.02)
+            except Empty:
+                break
+
+    def issue_command(self, cmd, add_history=False, catch_exception=True):
+        """Issues cmd to the persistent process and yields stdout and stderr messages.
 
         Each message is a dictionary with two keys:
 
-            - "type": either "stdout" or "stderr"
+            - "type": message type, e.g. "stdout" or "stderr"
             - "data": the actual message string.
 
         Args:
             cmd (str)
+            add_history (bool)
 
-        Returns:
-            generator
+        Yields:
+            dict: message
         """
-        t = Thread(target=self._issue_command_and_wait_for_idle, args=(cmd, add_history))
+        cmd = self.make_complete_command(cmd)
+        if cmd is None:
+            return
+        t = threading.Thread(target=self._issue_command_and_wait_for_idle, args=(cmd, add_history, catch_exception))
         t.start()
         while True:
             msg = self._msg_queue.get()
-            yield msg
             if msg["type"] == "command_finished":
                 break
+            yield msg
         t.join()
 
-    def _make_complete_command(self):
-        complete_cmd = os.linesep.join(self._command_buffer)
-        if len(self._command_buffer) == 1:
-            complete_cmd += os.linesep
-        return complete_cmd
-
-    def _issue_command_and_wait_for_idle(self, cmd, add_history):
+    def _issue_command_and_wait_for_idle(self, cmd, add_history, catch_exception):
         """Issues command and wait for idle.
 
         Args:
@@ -154,32 +196,28 @@ class PersistentManagerBase:
         """
         with self._lock:
             self._msg_queue.put({"type": "stdin", "data": cmd})
-            self._command_buffer.append(cmd)
-            complete_cmd = self._make_complete_command()
-            is_complete = self._is_complete(complete_cmd)
-            self.command_successful = self._issue_command(cmd, is_complete, add_history)
-            if self.command_successful and is_complete:
+            self.command_successful = self._issue_command(cmd, catch_exception=catch_exception)
+            if self.command_successful:
+                if add_history:
+                    self._communicate("add_history", cmd, receive=False)
                 self.command_successful &= self._wait()
-                self._command_buffer.clear()
-        self._msg_queue.put({"type": "command_finished", "is_complete": is_complete})
+        self._msg_queue.put({"type": "command_finished"})
 
-    def _issue_command(self, cmd, is_complete=True, add_history=False):
+    def _issue_command(self, cmd, catch_exception=True):
         """Writes command to the process's stdin and flushes.
 
         Args:
             cmd (str): Command to pass to the persistent process
-            is_complete (bool): Whether or not the command is complete
-            add_history (bool): Whether or not to add the command to history
+
+        Returns:
+            bool: True if command was issued successfully, False otherwise
         """
+        if catch_exception:
+            cmd = self._catch_exception_command(cmd) if cmd else ""
         cmd += os.linesep
-        if is_complete:
-            cmd += os.linesep
         self._persistent.stdin.write(cmd.encode("UTF8"))
         try:
             self._persistent.stdin.flush()
-            if is_complete and add_history:
-                complete_cmd = self._make_complete_command()
-                self._communicate("add_history", complete_cmd, receive=False)
             return True
         except BrokenPipeError:
             return False
@@ -187,7 +225,7 @@ class PersistentManagerBase:
     def _wait(self):
         """Waits for the persistent process to become idle.
 
-        This is implemented by writing the sentinel to stdin and waiting for the _SENTINEL to come out of stdout.
+        This is implemented by pinging a server and waiting for the server to receive the ping.
 
         Returns:
             bool
@@ -196,28 +234,43 @@ class PersistentManagerBase:
         with socketserver.TCPServer((host, 0), None) as s:
             port = s.server_address[1]
         queue = Queue()
-        thread = Thread(target=self._listen_and_enqueue, args=(host, port, queue))
+        thread = threading.Thread(target=self._wait_ping, args=(host, port, queue))
         thread.start()
         queue.get()  # This blocks until the server is listening
-        sentinel = self._sentinel_command(host, port)
-        if not self._issue_command(sentinel):
+        ping = self._ping_command(host, port)
+        if not self._issue_command(ping, catch_exception=False):
             thread.join()
             return False
-        queue.get()
+        result = queue.get()  # This blocks until the server is done
         thread.join()
-        return True
+        if not self.is_persistent_alive():
+            self._msg_queue.put({"type": "stdout", "data": "Kernel died (×_×)"})
+            return self._persistent.returncode == 0
+        return result == "ok"
 
-    @staticmethod
-    def _listen_and_enqueue(host, port, queue):
-        """Listens on the server and enqueues all data received."""
-        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
-            s.bind((host, port))
-            s.listen()
+    def _wait_ping(self, host, port, queue, timeout=1):
+        """Listens on a server until a connection is made (ping) or the underlying process is found dead.
+        Puts None in the queue when the server starts listening and when it's done."""
+        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as server:
+            server.settimeout(timeout)
+            server.bind((host, port))
+            server.listen()
             queue.put(None)
-            conn, _ = s.accept()
+            while True:
+                try:
+                    conn, _ = server.accept()
+                    break
+                except socket.timeout:
+                    if not self.is_persistent_alive():
+                        queue.put("error")
+                        return
+            conn.settimeout(timeout)
             with conn:
                 while True:
-                    data = conn.recv(1024)
+                    try:
+                        data = conn.recv(1024)
+                    except socket.timeout:
+                        continue
                     queue.put(data.decode("UTF8", "replace"))
                     if not data:
                         break
@@ -280,17 +333,27 @@ class PersistentManagerBase:
         self._persistent.wait()
         self._persistent.stdout = os.devnull
         self._persistent.stderr = os.devnull
+        self.set_running_until_completion(False)
         self._start_persistent()
+        self._wait()
+        yield from self.drain_queue()
 
     def interrupt_persistent(self):
         """Interrupts the persistent process."""
-        Thread(target=self._do_interrupt_persistent).start()
+        threading.Thread(target=self._do_interrupt_persistent).start()
+        self._wait()
 
     def _do_interrupt_persistent(self):
-        sig = signal.CTRL_C_EVENT if sys.platform == "win32" else signal.SIGINT
-        for _ in range(3):
-            self._persistent.send_signal(sig)
-            time.sleep(1)
+        persistent = self._persistent  # Make local copy; other threads may set self._persistent to None while sleeping.
+        if persistent is None:
+            return
+        if sys.platform == "win32":
+            p = Process(target=_send_ctrl_c, args=(persistent.pid,))
+            p.start()
+            p.join()
+        else:
+            persistent.send_signal(signal.SIGINT)
+        self.set_running_until_completion(False)
 
     def is_persistent_alive(self):
         """Whether or not the persistent is still alive and ready to receive commands.
@@ -298,11 +361,28 @@ class PersistentManagerBase:
         Returns:
             bool
         """
-        return self._persistent.returncode is None
+        return self._persistent is not None and self._persistent.poll() is None
 
     def kill_process(self):
+        if self._persistent is None:
+            return
         self._persistent.kill()
-        self._persistent.communicate(timeout=5)
+        self._persistent.wait()
+        self._persistent.stdin.close()
+        self._persistent.stdout.close()
+        self._persistent.stderr.close()
+        self._persistent = None
+        self.set_running_until_completion(False)
+        persistent_process_semaphore.release()
+
+
+def _send_ctrl_c(pid):
+    kernel = ctypes.windll.kernel32
+    kernel.FreeConsole()
+    kernel.AttachConsole(pid)
+    kernel.SetConsoleCtrlHandler(None, 1)
+    kernel.GenerateConsoleCtrlEvent(0, 0)
+    sys.exit(0)
 
 
 class JuliaPersistentManager(PersistentManagerBase):
@@ -315,11 +395,21 @@ class JuliaPersistentManager(PersistentManagerBase):
         """See base class."""
         path = os.path.join(os.path.dirname(__file__), "spine_repl.jl").replace(os.sep, "/")
         host, port = self._server_address
-        return ["-i", "-e", f'include("{path}"); SpineREPL.start_server("{host}", {port})']
+        return [
+            "-i",
+            "-e",
+            f'include("{path}"); SpineREPL.start_server("{host}", {port})',
+            "--color=yes",
+            "--banner=yes",
+        ]
 
     @staticmethod
-    def _sentinel_command(host, port):
-        return f'SpineREPL.send_sentinel("{host}", {port})'
+    def _catch_exception_command(cmd):
+        return f"try SpineREPL.set_exception(false); @eval {cmd} catch; SpineREPL.set_exception(true); rethrow() end"
+
+    @staticmethod
+    def _ping_command(host, port):
+        return f'SpineREPL.ping("{host}", {port})'
 
 
 class PythonPersistentManager(PersistentManagerBase):
@@ -332,60 +422,97 @@ class PythonPersistentManager(PersistentManagerBase):
         """See base class."""
         path = os.path.dirname(__file__).replace(os.sep, "/")
         return [
-            "-q",
             "-i",
+            "-u",
             "-c",
             f"import sys; sys.ps1 = sys.ps2 = ''; sys.path.append('{path}'); "
             f"import spine_repl; spine_repl.start_server({self._server_address})",
         ]
 
     @staticmethod
-    def _sentinel_command(host, port):
-        return f'spine_repl.send_sentinel("{host}", {port})'
+    def _catch_exception_command(cmd):
+        cmd_lines = cmd.splitlines()
+        indent = "\t" if any(l.startswith("\t") for l in cmd_lines) else "  "
+        lines = ["try:"]
+        lines += [indent + "spine_repl.set_exception(False)"]
+        for l in cmd_lines:
+            lines += [indent + l]
+        lines += ["except:"]
+        lines += [indent + "spine_repl.set_exception(True)"]
+        lines += [indent + "raise"]
+        return os.linesep.join(lines) + os.linesep
+
+    @staticmethod
+    def _ping_command(host, port):
+        return f'spine_repl.ping("{host}", {port})'
 
 
 class _PersistentManagerFactory(metaclass=Singleton):
-    _persistent_managers = {}
+    @dataclass
+    class OpenSign:
+        open: bool = True
+
+        def __bool__(self):
+            return self.open
+
+    persistent_managers = {}
     """Maps tuples (process args) to associated PersistentManagerBase."""
+    _isolated_managers = []
+    factory_lock = threading.Lock()
+    _factory_open = OpenSign()
 
     def new_persistent_manager(self, constructor, logger, args, group_id):
         """Creates a new persistent for given args and group id if none exists.
 
         Args:
-            constructor (function): the persistent manager constructor
+            constructor (Callable): the persistent manager constructor
             logger (LoggerInterface)
             args (list): the arguments to launch the persistent process
             group_id (str): item group that will execute using this persistent
 
         Returns:
-            PersistentManagerBase: persistent manager
+            PersistentManagerBase: persistent manager or None if factory has been closed
         """
-        if group_id is None:
-            # Execute in isolation
-            return constructor(args)
-        key = tuple(args + [group_id])
-        if key not in self._persistent_managers or not self._persistent_managers[key].is_persistent_alive():
-            try:
-                self._persistent_managers[key] = constructor(args)
-            except OSError as err:
-                msg = dict(type="persistent_failed_to_start", args=" ".join(args), error=str(err))
+        with self.factory_lock:
+            if group_id is None:
+                # Execute in isolation
+                with acquire_persistent_process(self.persistent_managers, self._isolated_managers, self._factory_open):
+                    if not self._factory_open:
+                        return None
+                    mp = constructor(args)
+                    self._isolated_managers.append(mp)
+                    return mp
+            key = tuple(args + [group_id])
+            if key not in self.persistent_managers or not self.persistent_managers[key].is_persistent_alive():
+                with acquire_persistent_process(self.persistent_managers, self._isolated_managers, self._factory_open):
+                    if not self._factory_open:
+                        return None
+                    try:
+                        self.persistent_managers[key] = pm = constructor(args)
+                    except OSError as err:
+                        msg = dict(type="persistent_failed_to_start", args=" ".join(args), error=str(err))
+                        logger.msg_persistent_execution.emit(msg)
+                        return None
+            else:
+                pm = self.persistent_managers[key]
+            msg = dict(type="persistent_started", key=key, language=pm.language)
+            logger.msg_persistent_execution.emit(msg)
+            for msg in pm.drain_queue():
                 logger.msg_persistent_execution.emit(msg)
-                return None
-        pm = self._persistent_managers[key]
-        msg = dict(type="persistent_started", key=key, language=pm.language)
-        logger.msg_persistent_execution.emit(msg)
-        return pm
+            return pm
 
     def restart_persistent(self, key):
         """Restart a persistent process.
 
         Args:
             key (tuple): persistent identifier
+        Returns:
+            generator: stdout and stderr messages (dictionaries with two keys: type, and data)
         """
-        pm = self._persistent_managers.get(key)
+        pm = self.persistent_managers.get(key)
         if pm is None:
-            return
-        pm.restart_persistent()
+            return ()
+        yield from pm.restart_persistent()
 
     def interrupt_persistent(self, key):
         """Interrupts a persistent process.
@@ -393,7 +520,7 @@ class _PersistentManagerFactory(metaclass=Singleton):
         Args:
             key (tuple): persistent identifier
         """
-        pm = self._persistent_managers.get(key)
+        pm = self.persistent_managers.get(key)
         if pm is None:
             return
         pm.interrupt_persistent()
@@ -406,14 +533,29 @@ class _PersistentManagerFactory(metaclass=Singleton):
             cmd (str): command to issue
 
         Returns:
-            generator: stdio and stderr messages (dictionaries with two keys: type, and data)
+            generator: stdin, stdout, and stderr messages (dictionaries with two keys: type, and data)
         """
-        pm = self._persistent_managers.get(key)
+        pm = self.persistent_managers.get(key)
         if pm is None:
-            return
-        for msg in pm.issue_command(cmd, add_history=True):
+            return ()
+        for msg in pm.issue_command(cmd, add_history=True, catch_exception=False):
             yield msg
 
+    def is_persistent_command_complete(self, key, cmd):
+        """Checkes whether a command is complete.
+
+        Args:
+            key (tuple): persistent identifier
+            cmd (str): command to issue
+
+        Returns:
+            bool
+        """
+        pm = self.persistent_managers.get(key)
+        if pm is None:
+            return False
+        return pm.make_complete_command(cmd) is not None
+
     def get_persistent_completions(self, key, text):
         """Returns a list of completion options.
 
@@ -424,38 +566,52 @@ class _PersistentManagerFactory(metaclass=Singleton):
         Returns:
             list of str: options that match given text
         """
-        pm = self._persistent_managers.get(key)
+        pm = self.persistent_managers.get(key)
         if pm is None:
             return
         return pm.get_completions(text)
 
     def get_persistent_history_item(self, key, index):
-        """Issues a command to a persistent process.
+        """Returns a history item.
 
         Args:
             key (tuple): persistent identifier
-            index (int): index of the history item, most recen first
+            index (int): index of the history item, most recent first
 
         Returns:
             str: history item or empty string if none
         """
-        pm = self._persistent_managers.get(key)
+        pm = self.persistent_managers.get(key)
         if pm is None:
             return
         return pm.get_history_item(index)
 
     def kill_manager_processes(self):
         """Kills persistent managers' Popen instances."""
-        for manager in self._persistent_managers.values():
+        for manager in chain(self.persistent_managers.values(), self._isolated_managers):
             manager.kill_process()
 
+    def open_factory(self):
+        """Opens factory.
+
+        An open factory instantiates new persistent managers.
+        """
+        self._factory_open.open = True
+
+    def close_factory(self):
+        """Closes the factory.
+
+        A closed factory does not instantiate new persistent managers
+        """
+        self._factory_open.open = False
+
 
 _persistent_manager_factory = _PersistentManagerFactory()
 
 
 def restart_persistent(key):
     """See _PersistentManagerFactory."""
-    _persistent_manager_factory.restart_persistent(key)
+    yield from _persistent_manager_factory.restart_persistent(key)
 
 
 def interrupt_persistent(key):
@@ -472,11 +628,26 @@ def kill_persistent_processes():
     _persistent_manager_factory.kill_manager_processes()
 
 
+def enable_persistent_process_creation():
+    """Enables creation of new persistent project managers."""
+    _persistent_manager_factory.open_factory()
+
+
+def disable_persistent_process_creation():
+    """Disables creation of new persistent project managers."""
+    _persistent_manager_factory.close_factory()
+
+
 def issue_persistent_command(key, cmd):
     """See _PersistentManagerFactory."""
     yield from _persistent_manager_factory.issue_persistent_command(key, cmd)
 
 
+def is_persistent_command_complete(key, cmd):
+    """See _PersistentManagerFactory."""
+    return _persistent_manager_factory.is_persistent_command_complete(key, cmd)
+
+
 def get_persistent_completions(key, text):
     """See _PersistentManagerFactory."""
     return _persistent_manager_factory.get_persistent_completions(key, text)
@@ -487,16 +658,36 @@ def get_persistent_history_item(key, index):
     return _persistent_manager_factory.get_persistent_history_item(key, index)
 
 
+@contextlib.contextmanager
+def acquire_persistent_process(group_persistent_managers, isolated_persistent_managers, running):
+    while running and not persistent_process_semaphore.acquire(timeout=0.5):
+        if not running:
+            break
+        killed = False
+        for pm in group_persistent_managers.values():
+            if pm.is_persistent_alive() and not pm.is_running_until_completion():
+                pm.kill_process()
+                killed = True
+                break
+        if not killed:
+            for i, pm in enumerate(isolated_persistent_managers):
+                if pm.is_persistent_alive() and pm.is_running_until_completion():
+                    pm.kill_process()
+                    isolated_persistent_managers.pop(i)
+                    break
+    yield None
+
+
 class PersistentExecutionManagerBase(ExecutionManagerBase):
     """Base class for managing execution of commands on a persistent process."""
 
     def __init__(self, logger, args, commands, alias, group_id=None):
-        """Class constructor.
-
+        """
         Args:
             logger (LoggerInterface): a logger instance
             args (list): List of args to start the persistent process
             commands (list): List of commands to execute in the persistent process
+            alias (str): an alias name for the manager
             group_id (str, optional): item group that will execute using this kernel
         """
         super().__init__(logger)
@@ -517,29 +708,35 @@ class PersistentExecutionManagerBase(ExecutionManagerBase):
         (a subclass of PersistentManagerBase)
 
         Returns:
-            function
+            Callable
         """
         raise NotImplementedError()
 
     def run_until_complete(self):
         """See base class."""
-        msg = dict(type="execution_started", args=" ".join(self._args))
-        self._logger.msg_persistent_execution.emit(msg)
-        self._logger.msg_persistent_execution.emit(dict(type="stdin", data=self._alias.strip()))
-        failed = False
-        for cmd in self._commands:
-            for msg in self._persistent_manager.issue_command(cmd):
-                if msg["type"] != "stdin":
-                    self._logger.msg_persistent_execution.emit(msg)
-                    if msg["type"] in ("stdout", "stderr"):
-                        failed = msg["type"] == "stderr"
-            if not self._persistent_manager.command_successful:
-                return -1
-        return -1 if failed else 0
+        if self._persistent_manager is None:
+            return -1
+        self._persistent_manager.set_running_until_completion(True)
+        try:
+            msg = dict(type="execution_started", args=" ".join(self._args))
+            self._logger.msg_persistent_execution.emit(msg)
+            fmt_alias = "# Running " + self._alias.rstrip()
+            self._logger.msg_persistent_execution.emit(dict(type="stdin", data=fmt_alias))
+            for cmd in self._commands:
+                for msg in self._persistent_manager.issue_command(cmd):
+                    if msg["type"] != "stdin":
+                        self._logger.msg_persistent_execution.emit(msg)
+                self.killed = not self._persistent_manager.is_persistent_alive()
+                if not self._persistent_manager.command_successful:
+                    return -1
+            return 0
+        finally:
+            self._persistent_manager.set_running_until_completion(False)
 
     def stop_execution(self):
         """See base class."""
-        self._persistent_manager.interrupt_persistent()
+        if self._persistent_manager is not None:
+            self._persistent_manager.interrupt_persistent()
 
 
 class JuliaPersistentExecutionManager(PersistentExecutionManagerBase):
