@@ -20,6 +20,7 @@ import sys
 import subprocess
 from threading import Thread
 from .execution_manager_base import ExecutionManagerBase
+from ..utils.execution_resources import one_shot_process_semaphore
 
 
 class ProcessExecutionManager(ExecutionManagerBase):
@@ -29,35 +30,41 @@ class ProcessExecutionManager(ExecutionManagerBase):
         Args:
             logger (LoggerInterface): a logger instance
             program (str): Path to program to run in the subprocess (e.g. julia.exe)
-            args (list): List of argument for the program (e.g. path to script file)
+            *args: List of arguments for the program (e.g. path to script file)
         """
         super().__init__(logger)
         self._process = None
         self._program = program
         self._args = args
         self._workdir = workdir
+        self._stopped = False
 
     def run_until_complete(self):
+        self._stopped = False
         cf = subprocess.CREATE_NO_WINDOW if sys.platform == "win32" else 0  # Don't show console when frozen
-        try:
-            self._process = subprocess.Popen(
-                [self._program, *self._args],
-                stdout=subprocess.PIPE,
-                stderr=subprocess.PIPE,
-                cwd=self._workdir,
-                creationflags=cf,
-            )
-        except OSError as e:
-            msg = dict(type="execution_failed_to_start", error=str(e), program=self._program)
+        with one_shot_process_semaphore:
+            if self._stopped:
+                return 0
+            try:
+                self._process = subprocess.Popen(
+                    [self._program, *self._args],
+                    stdout=subprocess.PIPE,
+                    stderr=subprocess.PIPE,
+                    cwd=self._workdir,
+                    creationflags=cf,
+                )
+            except OSError as e:
+                msg = dict(type="execution_failed_to_start", error=str(e), program=self._program)
+                self._logger.msg_standard_execution.emit(msg)
+                return 1
+            msg = dict(type="execution_started", program=self._program, args=" ".join(self._args))
             self._logger.msg_standard_execution.emit(msg)
-            return
-        msg = dict(type="execution_started", program=self._program, args=" ".join(self._args))
-        self._logger.msg_standard_execution.emit(msg)
-        Thread(target=self._log_stdout, args=(self._process.stdout,), daemon=True).start()
-        Thread(target=self._log_stderr, args=(self._process.stderr,), daemon=True).start()
-        return self._process.wait()
+            Thread(target=self._log_stdout, args=(self._process.stdout,), daemon=True).start()
+            Thread(target=self._log_stderr, args=(self._process.stderr,), daemon=True).start()
+            return self._process.wait()
 
     def stop_execution(self):
+        self._stopped = True
         if self._process is not None:
             self._process.terminate()
 
